@@ -1,17 +1,11 @@
-#!/usr/bin/env python3
-"""Script and library for symbolizing MongoDB stack traces.
+"""Script for symbolizing MongoDB stack traces.
 
 To use as a script, paste the JSON object on the line after ----- BEGIN BACKTRACE ----- into the
-standard input of this script. There are numerous caveats. In the default mode, you need
-to pass in the path to the executable being symbolized, and if you want shared library stack
-traces, you must be on the same system.
-
-There is largely untested support for extracting debug information from S3 buckets. This work
-is experimental.
+standard input of this script.
 
 Sample usage:
 
-db-contrib-tool symbolize --symbolizer-path=/path/to/llvm-symbolizer /path/to/executable < /file/with/stacktrace
+db-contrib-tool symbolize --symbolizer-path=/path/to/llvm-symbolizer < /file/with/stacktrace
 or more simply:
 db-contrib-tool symbolize < file/with/stackraces
 
@@ -26,11 +20,16 @@ import subprocess
 import sys
 import time
 from collections import OrderedDict
-from typing import Dict
+from enum import Enum
+from typing import Any, Dict, List, NamedTuple
 
+import inject
 import requests
+from tenacity import Retrying, retry_if_result, stop_after_delay, wait_fixed
 
+from db_contrib_tool.clients.io_client import IOClient
 from db_contrib_tool.plugin import PluginInterface, Subcommand, SubcommandResult
+from db_contrib_tool.symbolizer.mongo_log_parser_service import MongoLogParserService
 from db_contrib_tool.utils.build_system_options import PathOptions
 from db_contrib_tool.utils.oauth import (
     Configs,
@@ -358,33 +357,80 @@ class PathResolver(object):
         return os.path.join(path, inner_folder_name, binary_name)
 
 
-class Symbolizer(Subcommand):
-    def __init__(
-        self,
-        output_format,
-        host,
-        cache_dir,
-        input_format,
-        src_dir_to_move,
-        dsym_hint,
-        symbolizer_path,
-        live,
-        client_secret,
-        client_id,
-    ):
-        self.output_format = output_format
-        self.host = host
-        self.cache_dir = cache_dir
-        self.input_format = input_format
-        self.src_dir_to_move = src_dir_to_move
-        self.dsym_hint = dsym_hint
-        self.symbolizer_path = symbolizer_path
-        self.live = live
-        self.client_secret = client_secret
-        self.client_id = client_id
+class InputFormat(str, Enum):
+    """Input mongo log format types."""
 
-    def substitute_stdin(self, resolver):
-        """Accept stdin stream as source of logs and symbolize it."""
+    CLASSIC = "classic"
+    THIN = "thin"
+
+
+class OutputFormat(str, Enum):
+    """Output format types."""
+
+    CLASSIC = "classic"
+    JSON = "json"
+
+
+class SymbolizerParameters(NamedTuple):
+    """
+    Parameters describing how logs should be symbolized.
+
+    * symbolizer_path: Symbolizer executable path.
+    * dsym_hint: List of `-dsym-hint` flag values to pass to symbolizer.
+
+    * input_format: Input mongo log format.
+    * output_format: Output format.
+    * src_dir_to_move: Src dir to move to /data/mci/{original_buildid}/src.
+
+    * live: Whether it should enter live mode.
+
+    * host: URL of web service running the API to get debug symbol URL.
+    * cache_dir: Full path to a directory to store cache/files.
+    * total_seconds_for_retries: Timeout for getting data from web service.
+
+    * client_secret: Secret key for Okta Oauth.
+    * client_id: Client id for Okta Oauth.
+    """
+
+    symbolizer_path: str
+    dsym_hint: List[str]
+
+    input_format: InputFormat
+    output_format: OutputFormat
+    src_dir_to_move: str
+
+    live: bool
+
+    host: str
+    cache_dir: str
+    total_seconds_for_retries: int
+
+    client_secret: str
+    client_id: str
+
+
+class SymbolizerOrchestrator:
+    """Orchestrator for log symbolizer."""
+
+    @inject.autoparams()
+    def __init__(
+        self, mongo_log_parser_service: MongoLogParserService, io_client: IOClient
+    ) -> None:
+        """
+        Initialize.
+
+        :param mongo_log_parser_service: Service to parse backtraces in log files.
+        """
+        self.mongo_log_parser_service = mongo_log_parser_service
+        self.io_client = io_client
+
+    def substitute_stdin(self, resolver: PathResolver, params: SymbolizerParameters) -> None:
+        """
+        Accept stdin stream as source of logs and symbolize it.
+
+        :param resolver: Debug symbols path resolver.
+        :param params: Symbolizer parameters.
+        """
 
         # Ignore Ctrl-C. When the process feeding the pipe exits, `stdin` will be closed.
         signal.signal(signal.SIGINT, signal.SIG_IGN)
@@ -410,7 +456,12 @@ class Symbolizer(Subcommand):
                     print("Trace is empty, skipping...")
                     continue
                 frames = self.symbolize_frames(
-                    trace_doc, resolver, self.symbolizer_path, [], self.output_format
+                    trace_doc=trace_doc,
+                    dbg_path_resolver=resolver,
+                    symbolizer_path=params.symbolizer_path,
+                    dsym_hint=[],
+                    input_format=params.input_format,
+                    total_seconds_for_retries=params.total_seconds_for_retries,
                 )
                 print(prefix)
                 print("Symbolizing...")
@@ -419,71 +470,50 @@ class Symbolizer(Subcommand):
             else:
                 print(line)
 
-    def execute(self):
+    def execute(self, params: SymbolizerParameters) -> bool:
+        """
+        Execute symbolizer.
+
+        :param params: Symbolizer parameters.
+        :return: Succeeded or not.
+        """
         resolver = PathResolver(
-            host=self.host,
-            cache_dir=self.cache_dir,
-            client_secret=self.client_secret,
-            client_id=self.client_id,
+            host=params.host,
+            cache_dir=params.cache_dir,
+            client_secret=params.client_secret,
+            client_id=params.client_id,
         )
 
-        if self.live:
+        if params.live:
             print("Entering live mode")
-            self.substitute_stdin(resolver)
-            return SubcommandResult.SUCCESS
+            self.substitute_stdin(resolver, params)
+            return True
 
-        # Skip over everything before the first '{' since it is likely to be log line prefixes.
-        # Additionally, using raw_decode() to ignore extra data after the closing '}' to allow maximal
-        # sloppiness in copy-pasting input.
-        trace_doc = sys.stdin.read()
-
-        if not trace_doc or not trace_doc.strip():
+        log_lines = self.io_client.read_from_stdin()
+        if not log_lines or not log_lines.strip():
             print(
                 "Please provide the backtrace through stdin for symbolization;"
                 "e.g. `your/symbolization/command < /file/with/stacktrace`"
             )
 
-        # Search the trace_doc for an object having "backtrace" and "processInfo" keys.
-        def bt_search(obj):
-            try:
-                if "backtrace" in obj and "processInfo" in obj:
-                    return obj
-                for _, val in obj.items():
-                    res = bt_search(val)
-                    if res:
-                        return res
-            except (TypeError, AttributeError):
-                pass
-            return None
-
-        # given a log file including traceback,
-        # we try to find traceback from that file, analyzing each line until we find it
-        for line in trace_doc.splitlines():
-            possible_trace_doc = line[line.find("{") :]
-            try:
-                possible_trace_doc = json.JSONDecoder().raw_decode(possible_trace_doc)[0]
-                trace_doc = bt_search(possible_trace_doc)
-                if trace_doc:
-                    break
-            except json.JSONDecodeError:
-                pass
-        else:
+        trace_doc = self.mongo_log_parser_service.parse_log_lines(log_lines)
+        if trace_doc is None:
             print("could not find json backtrace object in input", file=sys.stderr)
-            return SubcommandResult.FAIL
+            return False
 
         output_fn = None
-        if self.output_format == "json":
+        if params.output_format == OutputFormat.JSON:
             output_fn = json.dump
-        if self.output_format == "classic":
+        if params.output_format == OutputFormat.CLASSIC:
             output_fn = self.classic_output
 
-        frames = self.preprocess_frames(resolver, trace_doc, self.input_format)
+        frames = self.preprocess_frames_with_retries(resolver, trace_doc, params.input_format)
 
-        if self.src_dir_to_move and resolver.mci_build_dir is not None:
+        if params.src_dir_to_move and resolver.mci_build_dir is not None:
             try:
                 os.makedirs(resolver.mci_build_dir)
                 os.symlink(
-                    os.path.join(os.getcwd(), self.src_dir_to_move),
+                    os.path.join(os.getcwd(), params.src_dir_to_move),
                     os.path.join(resolver.mci_build_dir, "src"),
                 )
             except FileExistsError:
@@ -493,13 +523,14 @@ class Symbolizer(Subcommand):
             frames=frames,
             trace_doc=trace_doc,
             dbg_path_resolver=resolver,
-            symbolizer_path=self.symbolizer_path,
-            dsym_hint=self.dsym_hint,
-            input_format=self.input_format,
+            symbolizer_path=params.symbolizer_path,
+            dsym_hint=params.dsym_hint,
+            input_format=params.input_format,
+            total_seconds_for_retries=params.total_seconds_for_retries,
         )
         output_fn(frames, sys.stdout, indent=2)
 
-        return SubcommandResult.SUCCESS
+        return True
 
     @staticmethod
     def parse_input(trace_doc, dbg_path_resolver):
@@ -561,7 +592,7 @@ class Symbolizer(Subcommand):
         frames = kwargs.get("frames")
         if frames is None:
             total_seconds_for_retries = kwargs.get("total_seconds_for_retries", 0)
-            frames = self.preprocess_frames(
+            frames = self.preprocess_frames_with_retries(
                 dbg_path_resolver,
                 trace_doc,
                 input_format,
@@ -631,86 +662,62 @@ class Symbolizer(Subcommand):
         return frames
 
     def preprocess_frames(
-        self,
-        dbg_path_resolver,
-        trace_doc,
-        input_format,
-        total_seconds_for_retries=0,
-        current_retry_number=0,
-    ):
+        self, dbg_path_resolver: PathResolver, trace_doc: Dict[str, Any], input_format: InputFormat
+    ) -> List[Dict[str, Any]]:
         """
         Process the paths in frame objects.
 
-        :param dbg_path_resolver: resolver class
-        :param trace_doc: traceback string
-        :param input_format: format of input
-        :param total_seconds_for_retries: max wait time for retries in seconds
-        :param current_retry_number: current retry counter
+        :param dbg_path_resolver: Debug symbols path resolver.
+        :param trace_doc: Traceback doc.
+        :param input_format: Input format.
+        :return: List of traceback frames.
         """
-        if input_format == "classic":
-            frames = self.parse_input(trace_doc, dbg_path_resolver)
-        elif input_format == "thin":
+
+        if input_format == InputFormat.CLASSIC:
+            return self.parse_input(trace_doc, dbg_path_resolver)
+        elif input_format == InputFormat.THIN:
             frames = trace_doc["backtrace"]
             for frame in frames:
                 frame["path"] = dbg_path_resolver.get_dbg_file(frame)
-        else:
-            raise ValueError('Unknown input format "{}"'.format(input_format))
+            return frames
 
-        # Check if path resolver actually found enough paths.
-        # If (not found / all) ratio is higher than 0.5, we retry.
-        not_found = [1 for i in frames if i.get("path") is None]
+    @staticmethod
+    def has_high_not_found_paths_ratio(frames: List[Dict[str, Any]]) -> bool:
+        """
+        Check whether not found paths in frames ratio is higher than 0.5.
+
+        :param frames: the list of traceback frames
+        :return: True if ratio is higher than 0.5
+        """
+        not_found = [1 for f in frames if f.get("path") is None]
         not_found_ratio = len(not_found) / (len(frames) or 1)
-        # When we fail to find more than 50% of paths (not_found_ratio >= 0.5),
-        # we wait for specific amount of time,
-        # and then retry again. this continues until we hit total_seconds_for_retries.
-        # After each wait, we increment next wait time, to not flood with too many retries with the same interval.
-        # There was thought about incrementing exponentially, by multiplying by 2 (time *= 2),
-        # but that step seemed too high, making less breakpoints to stop and retry.
-        # Another idea was to add 1 minute each time, but that felt like too frequent.
-        # The final decision was to go with 2: 2 extra minutes after each wait.
-        increment_mins = 2
+        return not_found_ratio >= 0.5
 
-        if not_found_ratio >= 0.5 and total_seconds_for_retries > 0:
-            # Retry processing.
-            # Our total wait time should not exceed total_seconds_for_retries.
-            # We should calculate how long we already waited up to this point.
-            # Since we increment wait time by 2 minutes each time,
-            # it makes a perfect arithmetic sequence.
-            # We can calculate the sum of sequence using this formula: S = n/2 (a + L)
-            # where: S - sum, n - number of items in sequence, a - first term, L - last term.
-            # Items in our sequence is equal to current_retry_number.
-            # First term is always 2, which is the increment_mins value.
-            # Last term is current_retry_number.
-            # Formula: S = current_retry_number/2 (2 + current_retry_number).
-            # Do not forget to multiply by 60 to convert into seconds.
-            total_wait_time = (
-                current_retry_number / 2 * (increment_mins + current_retry_number) * 60
-            )
-            if total_wait_time > total_seconds_for_retries:
-                # We already waited/tried enough.
-                return frames
+    def preprocess_frames_with_retries(
+        self,
+        dbg_path_resolver: PathResolver,
+        trace_doc: Dict[str, Any],
+        input_format: InputFormat,
+        total_seconds_for_retries: int = 0,
+    ) -> List[Dict[str, Any]]:
+        """
+        Process the paths in frame objects.
 
-            current_retry_number += 1
-            # We can multiply current_retry_number by increment_mins value,
-            # and it will be the same as current_retry_number += increment_mins
-            wait_time = current_retry_number * increment_mins * 60
-            if (total_wait_time + wait_time) > total_seconds_for_retries:
-                # If we wait for another wait_time, our total wait time will exceed total_seconds_for_retries.
-                # We don't want that. so, we justify it to the amount where it doesn't exceed the limit.
-                wait_time = total_seconds_for_retries - total_wait_time
-            print(
-                f"Currently {not_found_ratio * 100:.1f}% of frames couldn't be resolved. "
-                f"Retrying again in {wait_time} seconds..."
-            )
-            time.sleep(wait_time)
-            return self.preprocess_frames(
-                dbg_path_resolver,
-                trace_doc,
-                input_format,
-                total_seconds_for_retries,
-                current_retry_number,
-            )
-        return frames
+        :param dbg_path_resolver: Debug symbols path resolver.
+        :param trace_doc: Traceback doc.
+        :param input_format: Input format.
+        :param total_seconds_for_retries: Timeout for retries.
+        :return: List of traceback frames.
+        """
+
+        retrying = Retrying(
+            retry=retry_if_result(self.has_high_not_found_paths_ratio),
+            wait=wait_fixed(60),
+            stop=stop_after_delay(total_seconds_for_retries),
+            retry_error_callback=lambda retry_state: retry_state.outcome.result(),
+        )
+
+        return retrying(self.preprocess_frames, dbg_path_resolver, trace_doc, input_format)
 
     @staticmethod
     def classic_output(frames, outfile, **kwargs):
@@ -728,6 +735,50 @@ class Symbolizer(Subcommand):
                 )
 
 
+class Symbolizer(Subcommand):
+    """Main class for the symbolizer subcommand."""
+
+    def __init__(
+        self,
+        symbolizer_path: str,
+        dsym_hint: List[str],
+        input_format: str,
+        output_format: str,
+        src_dir_to_move: str,
+        live: bool,
+        host: str,
+        cache_dir: str,
+        total_seconds_for_retries: int,
+        client_secret: str,
+        client_id: str,
+    ):
+        """Initialize."""
+
+        self.symbolizer_params = SymbolizerParameters(
+            symbolizer_path=symbolizer_path,
+            dsym_hint=dsym_hint,
+            input_format=InputFormat.CLASSIC if input_format == "classic" else InputFormat.THIN,
+            output_format=OutputFormat.CLASSIC if output_format == "classic" else OutputFormat.JSON,
+            src_dir_to_move=src_dir_to_move,
+            live=live,
+            host=host,
+            cache_dir=cache_dir,
+            total_seconds_for_retries=total_seconds_for_retries,
+            client_secret=client_secret,
+            client_id=client_id,
+        )
+
+    def execute(self) -> SubcommandResult:
+        """Execute symbolize subcommand."""
+        inject.configure()
+        symbolize_orchestrator = inject.instance(SymbolizerOrchestrator)
+
+        success = symbolize_orchestrator.execute(self.symbolizer_params)
+        if success:
+            return SubcommandResult.SUCCESS
+        return SubcommandResult.FAIL
+
+
 class SymbolizerPlugin(PluginInterface):
     """CLI plugin for Symbolizer."""
 
@@ -738,23 +789,37 @@ class SymbolizerPlugin(PluginInterface):
             return None
 
         return Symbolizer(
+            symbolizer_path=parsed_args.symbolizer_path,
+            dsym_hint=parsed_args.dsym_hint,
+            input_format=parsed_args.input_format,
             output_format=parsed_args.output_format,
+            src_dir_to_move=parsed_args.src_dir_to_move,
+            live=parsed_args.live,
             host=parsed_args.host,
             cache_dir=parsed_args.cache_dir,
-            input_format=parsed_args.input_format,
-            src_dir_to_move=parsed_args.src_dir_to_move,
-            dsym_hint=parsed_args.dsym_hint,
-            symbolizer_path=parsed_args.symbolizer_path,
-            live=parsed_args.live,
+            total_seconds_for_retries=parsed_args.total_seconds_for_retries,
             client_secret=parsed_args.client_secret,
             client_id=parsed_args.client_id,
         )
 
     def add_subcommand(self, subparsers):
         parser = subparsers.add_parser(SUBCOMMAND, help=__doc__)
-        parser.add_argument("--dsym-hint", default=[], action="append")
-        parser.add_argument("--symbolizer-path", default="")
-        parser.add_argument("--input-format", choices=["classic", "thin"], default="classic")
+        parser.add_argument(
+            "--symbolizer-path", default="", help="Path to llvm-symbolizer executable."
+        )
+        parser.add_argument(
+            "--dsym-hint",
+            default=[],
+            action="append",
+            help="`-dsym-hint` flag values to pass to llvm-symbolizer.",
+        )
+
+        parser.add_argument(
+            "--input-format",
+            choices=["classic", "thin"],
+            default="classic",
+            help="Input mongo log format type.",
+        )
         parser.add_argument(
             "--output-format",
             choices=["classic", "json"],
@@ -769,27 +834,22 @@ class SymbolizerPlugin(PluginInterface):
             help="Specify a src dir to move to /data/mci/{original_buildid}/src",
         )
 
+        parser.add_argument("--live", action="store_true", help="Enter live mode.")
+
+        parser.add_argument(
+            "--host", default="", help="URL of web service running the API to get debug symbol URL"
+        )
+        # caching mechanism is currently not fully developed and needs more advanced cleaning techniques,
+        # we add an option to enable it after completing the implementation.
+        parser.add_argument(
+            "--cache-dir", default="", help="Full path to a directory to store cache/files"
+        )
         parser.add_argument(
             "--total-seconds-for-retries",
             default=0,
             type=int,
-            help="If web service fails to find path for given build id, it could be because mapping "
-            "process was not finished yet. We can wait for it to finish and retry again. Each retry"
-            " adds 2 minutes to previous wait time. It is guaranteed that total wait time does not exceed this "
-            "specified amount.",
+            help="Timeout for getting data from web service.",
         )
 
-        parser.add_argument("--live", action="store_true")
-        parser.add_argument(
-            "--host", default="", help="URL of web service running the API to get debug symbol URL"
-        )
-        parser.add_argument(
-            "--cache-dir", default="", help="Full path to a directory to store cache/files"
-        )
         parser.add_argument("--client-secret", default="", help="Secret key for Okta Oauth")
         parser.add_argument("--client-id", default="", help="Client id for Okta Oauth")
-        # caching mechanism is currently not fully developed and needs more advanced cleaning techniques,
-        # we add an option to enable it after completing the implementation.
-
-        # Look for symbols in the cwd by default.
-        parser.add_argument("path_to_executable", nargs="?")
